class SchemaCommandError(Exception):
    """
    Base class for all schema command errors.
    """

    pass


class GenerateError(SchemaCommandError):
    """
    Raised when an error occurs during schema generation.
    """

    pass



from tests.base import DBTestCase
from tests.example_apps.music.tables import Manager


class TestTableRepr(DBTestCase):
    def test_repr_postgres(self):
        self.assertEqual(
            Manager().__repr__(),
            "<Manager: None>",
        )

        self.insert_row()
        manager = Manager.objects().first().run_sync()
        self.assertEqual(manager.__repr__(), f"<Manager: {manager.id}>")



from tests.base import DBTestCase
from tests.example_apps.music.tables import Band


class TestRefresh(DBTestCase):
    def setUp(self):
        super().setUp()
        self.insert_rows()

    def test_refresh(self):
        """
        Make sure ``refresh`` works, with no columns specified.
        """
        # Fetch an instance from the database.
        band: Band = Band.objects().get(Band.name == "Pythonistas").run_sync()
        initial_data = band.to_dict()

        # Modify the data in the database.
        Band.update(
            {Band.name: Band.name + "!!!", Band.popularity: 8000}
        ).where(Band.name == "Pythonistas").run_sync()

        # Refresh `band`, and make sure it has the correct data.
        band.refresh().run_sync()

        self.assertTrue(band.name == "Pythonistas!!!")
        self.assertTrue(band.popularity == 8000)
        self.assertTrue(band.id == initial_data["id"])

    def test_columns(self):
        """
        Make sure ``refresh`` works, when columns are specified.
        """
        # Fetch an instance from the database.
        band: Band = Band.objects().get(Band.name == "Pythonistas").run_sync()
        initial_data = band.to_dict()

        # Modify the data in the database.
        Band.update(
            {Band.name: Band.name + "!!!", Band.popularity: 8000}
        ).where(Band.name == "Pythonistas").run_sync()

        # Refresh `band`, and make sure it has the correct data.
        query = band.refresh(columns=[Band.name])
        self.assertEqual(
            [i._meta.name for i in query._columns],
            ["name"],
        )
        query.run_sync()

        self.assertTrue(band.name == "Pythonistas!!!")
        self.assertTrue(band.popularity == initial_data["popularity"])
        self.assertTrue(band.id == initial_data["id"])

    def test_error_when_not_in_db(self):
        """
        Make sure we can't refresh an instance which hasn't been saved in the
        database.
        """
        band = Band()

        with self.assertRaises(ValueError) as manager:
            band.refresh().run_sync()

        self.assertEqual(
            "The instance doesn't exist in the database.",
            str(manager.exception),
        )

    def test_error_when_pk_in_none(self):
        """
        Make sure we can't refresh an instance when the primary key value isn't
        set.
        """
        band: Band = Band.objects().first().run_sync()
        band.id = None

        with self.assertRaises(ValueError) as manager:
            band.refresh().run_sync()

        self.assertEqual(
            "The instance's primary key value isn't defined.",
            str(manager.exception),
        )



import asyncio
import os
import tempfile
from unittest import TestCase
from unittest.mock import call, patch

from piccolo.engine.postgres import PostgresEngine
from piccolo.engine.sqlite import SQLiteEngine
from tests.base import DBTestCase, engine_is, engines_only, sqlite_only
from tests.example_apps.music.tables import Manager


@engines_only("postgres", "cockroach")
class TestPool(DBTestCase):
    async def _create_pool(self):
        engine: PostgresEngine = Manager._meta.db

        await engine.start_connection_pool()
        assert engine.pool is not None

        await engine.close_connection_pool()
        assert engine.pool is None

    async def _make_query(self):
        await Manager._meta.db.start_connection_pool()

        await Manager(name="Bob").save().run()
        response = await Manager.select().run()
        self.assertIn("Bob", [i["name"] for i in response])

        await Manager._meta.db.close_connection_pool()

    async def _make_many_queries(self):
        await Manager._meta.db.start_connection_pool()

        await Manager(name="Bob").save().run()

        async def get_data():
            response = await Manager.select().run()
            if engine_is("cockroach"):
                self.assertEqual(
                    response, [{"id": response[0]["id"], "name": "Bob"}]
                )
            else:
                self.assertEqual(response, [{"id": 1, "name": "Bob"}])

        await asyncio.gather(*[get_data() for _ in range(500)])

        await Manager._meta.db.close_connection_pool()

    def test_creation(self):
        """
        Make sure a connection pool can be created.
        """
        asyncio.run(self._create_pool())

    def test_query(self):
        """
        Make several queries using a connection pool.
        """
        asyncio.run(self._make_query())

    def test_many_queries(self):
        """
        Make sure the connection pool is working correctly, and we don't
        exceed a connection limit - queries should queue, then succeed.
        """
        asyncio.run(self._make_many_queries())


@engines_only("postgres", "cockroach")
class TestPoolProxyMethods(DBTestCase):
    async def _create_pool(self):
        engine: PostgresEngine = Manager._meta.db

        # Deliberate typo ('nnn'):
        await engine.start_connnection_pool()
        assert engine.pool is not None

        # Deliberate typo ('nnn'):
        await engine.close_connnection_pool()
        assert engine.pool is None

    def test_proxy_methods(self):
        """
        There are some proxy methods, due to some old typos. Make sure they
        work, to ensure backwards compatibility.
        """
        asyncio.run(self._create_pool())


@sqlite_only
class TestConnectionPoolWarning(TestCase):
    async def _create_pool(self):
        sqlite_file = os.path.join(tempfile.gettempdir(), "engine.sqlite")
        engine = SQLiteEngine(path=sqlite_file)

        with patch("piccolo.engine.base.colored_warning") as colored_warning:
            await engine.start_connection_pool()
            await engine.close_connection_pool()

            self.assertEqual(
                colored_warning.call_args_list,
                [
                    call(
                        "Connection pooling is not supported for sqlite.",
                        stacklevel=3,
                    ),
                    call(
                        "Connection pooling is not supported for sqlite.",
                        stacklevel=3,
                    ),
                ],
            )

    def test_warnings(self):
        """
        Make sure that when trying to start and close a connection pool with
        SQLite, a warning is printed out, as connection pools aren't currently
        supported.
        """
        asyncio.run(self._create_pool())



from unittest import TestCase
from unittest.mock import MagicMock

from piccolo.columns.column_types import Varchar
from piccolo.engine import engine_finder
from piccolo.engine.postgres import PostgresEngine
from piccolo.table import Table
from tests.base import AsyncMock, engines_only


@engines_only("postgres", "cockroach")
class TestExtraNodes(TestCase):
    def test_extra_nodes(self):
        """
        Make sure that other nodes can be queried.
        """
        # Get the test database credentials:
        test_engine = engine_finder()

        EXTRA_NODE = MagicMock(spec=PostgresEngine(config=test_engine.config))
        EXTRA_NODE.run_querystring = AsyncMock(return_value=[])

        DB = PostgresEngine(
            config=test_engine.config, extra_nodes={"read_1": EXTRA_NODE}
        )

        class Manager(Table, db=DB):
            name = Varchar()

        # Make sure the node is queried
        Manager.select().run_sync(node="read_1")
        self.assertTrue(EXTRA_NODE.run_querystring.called)

        # Make sure that a non existent node raises an error
        with self.assertRaises(KeyError):
            Manager.select().run_sync(node="read_2")



from unittest import TestCase

from tests.example_apps.music.tables import Band


class TestAllColumns(TestCase):
    def test_all_columns(self):
        self.assertEqual(
            Band.all_columns(),
            [Band.id, Band.name, Band.manager, Band.popularity],
        )
        self.assertEqual(Band.all_columns(), Band._meta.columns)

    def test_all_columns_excluding(self):
        self.assertEqual(
            Band.all_columns(exclude=[Band.id]),
            [Band.name, Band.manager, Band.popularity],
        )

        self.assertEqual(
            Band.all_columns(exclude=["id"]),
            [Band.name, Band.manager, Band.popularity],
        )



import typing as t
import uuid
from enum import Enum

from .base import Default


class UUID4(Default):
    @property
    def postgres(self):
        return "uuid_generate_v4()"

    @property
    def cockroach(self):
        return self.postgres

    @property
    def sqlite(self):
        return "''"

    def python(self):
        return uuid.uuid4()


UUIDArg = t.Union[UUID4, uuid.UUID, str, Enum, None]


__all__ = ["UUIDArg", "UUID4"]



import datetime
import decimal
import os
import tempfile
import typing as t
import uuid
from unittest import TestCase

from piccolo.apps.fixtures.commands.dump import (
    FixtureConfig,
    dump_to_json_string,
)
from piccolo.apps.fixtures.commands.load import load, load_json_string
from piccolo.utils.sync import run_sync
from tests.base import engines_only
from tests.example_apps.mega.tables import MegaTable, SmallTable


class TestDumpLoad(TestCase):
    """
    Test the fixture dump and load commands - makes sense to test them
    together.
    """

    maxDiff = None

    def setUp(self):
        for table_class in (SmallTable, MegaTable):
            table_class.create_table().run_sync()

    def tearDown(self):
        for table_class in (MegaTable, SmallTable):
            table_class.alter().drop_table().run_sync()

    def insert_rows(self):
        small_table = SmallTable(varchar_col="Test")
        small_table.save().run_sync()

        SmallTable(varchar_col="Test 2").save().run_sync()

        mega_table = MegaTable(
            bigint_col=1,
            boolean_col=True,
            bytea_col="hello".encode("utf8"),
            date_col=datetime.date(year=2021, month=1, day=1),
            foreignkey_col=small_table,
            integer_col=1,
            interval_col=datetime.timedelta(seconds=10),
            json_col={"a": 1},
            jsonb_col={"a": 1},
            numeric_col=decimal.Decimal("1.1"),
            real_col=1.1,
            double_precision_col=1.344,
            smallint_col=1,
            text_col="hello",
            timestamp_col=datetime.datetime(year=2021, month=1, day=1),
            timestamptz_col=datetime.datetime(
                year=2021, month=1, day=1, tzinfo=datetime.timezone.utc
            ),
            uuid_col=uuid.UUID("12783854-c012-4c15-8183-8eecb46f2c4e"),
            varchar_col="hello",
            unique_col="hello",
            null_col=None,
            not_null_col="hello",
        )
        mega_table.save().run_sync()

    def _run_comparison(self, table_class_names: t.List[str]):
        self.insert_rows()

        json_string = run_sync(
            dump_to_json_string(
                fixture_configs=[
                    FixtureConfig(
                        app_name="mega",
                        table_class_names=table_class_names,
                    )
                ]
            )
        )

        # We need to clear the data out now, otherwise when loading the data
        # back in, there will be constraint errors over clashing primary
        # keys.
        SmallTable.delete(force=True).run_sync()
        MegaTable.delete(force=True).run_sync()

        run_sync(load_json_string(json_string))

        self.assertEqual(
            SmallTable.select().run_sync(),
            [
                {"id": 1, "varchar_col": "Test"},
                {"id": 2, "varchar_col": "Test 2"},
            ],
        )

        mega_table_data = MegaTable.select().run_sync()

        # Real numbers don't have perfect precision when coming back from the
        # database, so we need to round them to be able to compare them.
        mega_table_data[0]["real_col"] = round(
            mega_table_data[0]["real_col"], 1
        )

        # Remove white space from the JSON values
        for col_name in ("json_col", "jsonb_col"):
            mega_table_data[0][col_name] = mega_table_data[0][
                col_name
            ].replace(" ", "")

        self.assertTrue(len(mega_table_data) == 1)

        self.assertDictEqual(
            mega_table_data[0],
            {
                "id": 1,
                "bigint_col": 1,
                "boolean_col": True,
                "bytea_col": b"hello",
                "date_col": datetime.date(2021, 1, 1),
                "foreignkey_col": 1,
                "integer_col": 1,
                "interval_col": datetime.timedelta(seconds=10),
                "json_col": '{"a":1}',
                "jsonb_col": '{"a":1}',
                "numeric_col": decimal.Decimal("1.1"),
                "real_col": 1.1,
                "double_precision_col": 1.344,
                "smallint_col": 1,
                "text_col": "hello",
                "timestamp_col": datetime.datetime(2021, 1, 1, 0, 0),
                "timestamptz_col": datetime.datetime(
                    2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
                ),
                "uuid_col": uuid.UUID("12783854-c012-4c15-8183-8eecb46f2c4e"),
                "varchar_col": "hello",
                "unique_col": "hello",
                "null_col": None,
                "not_null_col": "hello",
            },
        )

        # Make sure subsequent inserts work.
        SmallTable().save().run_sync()

    @engines_only("postgres", "sqlite")
    def test_dump_load(self):
        """
        Make sure we can dump some rows into a JSON fixture, then load them
        back into the database.
        """
        self._run_comparison(table_class_names=["SmallTable", "MegaTable"])

    @engines_only("postgres", "sqlite")
    def test_dump_load_ordering(self):
        """
        Similar to `test_dump_load` - but we need to make sure it inserts
        the data in the correct order, so foreign key constraints don't fail.
        """
        self._run_comparison(table_class_names=["MegaTable", "SmallTable"])

    @engines_only("cockroach")
    def test_dump_load_cockroach(self):
        """
        Similar to `test_dump_load`, except the schema is slightly different
        for CockroachDB.
        """
        self.insert_rows()

        json_string = run_sync(
            dump_to_json_string(
                fixture_configs=[
                    FixtureConfig(
                        app_name="mega",
                        table_class_names=["SmallTable", "MegaTable"],
                    )
                ]
            )
        )

        # We need to clear the data out now, otherwise when loading the data
        # back in, there will be constraint errors over clashing primary
        # keys.
        SmallTable.delete(force=True).run_sync()
        MegaTable.delete(force=True).run_sync()

        run_sync(load_json_string(json_string))

        result = SmallTable.select().run_sync()[0]
        result.pop("id")

        self.assertDictEqual(
            result,
            {"varchar_col": "Test"},
        )

        mega_table_data = MegaTable.select().run_sync()

        # Real numbers don't have perfect precision when coming back from the
        # database, so we need to round them to be able to compare them.
        mega_table_data[0]["real_col"] = round(
            mega_table_data[0]["real_col"], 1
        )

        # Remove white space from the JSON values
        for col_name in ("json_col", "jsonb_col"):
            mega_table_data[0][col_name] = mega_table_data[0][
                col_name
            ].replace(" ", "")

        self.assertTrue(len(mega_table_data) == 1)

        mega_table_data = mega_table_data[0]
        mega_table_data.pop("id")
        mega_table_data.pop("foreignkey_col")

        self.assertDictEqual(
            mega_table_data,
            {
                "bigint_col": 1,
                "boolean_col": True,
                "bytea_col": b"hello",
                "date_col": datetime.date(2021, 1, 1),
                "integer_col": 1,
                "interval_col": datetime.timedelta(seconds=10),
                "json_col": '{"a":1}',
                "jsonb_col": '{"a":1}',
                "numeric_col": decimal.Decimal("1.1"),
                "real_col": 1.1,
                "double_precision_col": 1.344,
                "smallint_col": 1,
                "text_col": "hello",
                "timestamp_col": datetime.datetime(2021, 1, 1, 0, 0),
                "timestamptz_col": datetime.datetime(
                    2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
                ),
                "uuid_col": uuid.UUID("12783854-c012-4c15-8183-8eecb46f2c4e"),
                "varchar_col": "hello",
                "unique_col": "hello",
                "null_col": None,
                "not_null_col": "hello",
            },
        )


class TestOnConflict(TestCase):
    def setUp(self) -> None:
        SmallTable.create_table().run_sync()
        SmallTable({SmallTable.varchar_col: "Test"}).save().run_sync()

    def tearDown(self) -> None:
        SmallTable.alter().drop_table().run_sync()

    def test_on_conflict(self):
        temp_dir = tempfile.gettempdir()

        json_file_path = os.path.join(temp_dir, "fixture.json")

        json_string = run_sync(
            dump_to_json_string(
                fixture_configs=[
                    FixtureConfig(
                        app_name="mega",
                        table_class_names=["SmallTable"],
                    )
                ]
            )
        )

        if os.path.exists(json_file_path):
            os.unlink(json_file_path)

        with open(json_file_path, "w") as f:
            f.write(json_string)

        run_sync(load(path=json_file_path, on_conflict="DO NOTHING"))
        run_sync(load(path=json_file_path, on_conflict="DO UPDATE"))
        run_sync(load(path=json_file_path, on_conflict="do nothing"))
        run_sync(load(path=json_file_path, on_conflict="do update"))



from unittest import TestCase

from piccolo.columns.column_types import Integer
from piccolo.table import Table
from tests.example_apps.music.tables import Manager


class TestIndexes(TestCase):
    def setUp(self):
        Manager.create_table().run_sync()

    def tearDown(self):
        Manager.alter().drop_table().run_sync()

    def test_create_index(self):
        """
        Test single column and multi column indexes.
        """
        for columns in [
            [Manager.name],
            [Manager._meta.primary_key, Manager.name],
        ]:
            Manager.create_index(columns).run_sync()

            index_name = Manager._get_index_name(
                [i._meta.name for i in columns]
            )

            index_names = Manager.indexes().run_sync()
            self.assertIn(index_name, index_names)

            Manager.drop_index(columns).run_sync()
            index_names = Manager.indexes().run_sync()
            self.assertNotIn(index_name, index_names)


class Concert(Table):
    order = Integer()


class TestProblematicColumnName(TestCase):
    def setUp(self):
        Concert.create_table().run_sync()

    def tearDown(self):
        Concert.alter().drop_table().run_sync()

    def test_problematic_name(self):
        """
        Make sure we can add an index to a column with a problematic name
        (which clashes with a SQL keyword).
        """
        columns = [Concert.order]
        Concert.create_index(columns=columns).run_sync()
        index_name = Concert._get_index_name([i._meta.name for i in columns])

        index_names = Concert.indexes().run_sync()
        self.assertIn(index_name, index_names)

        Concert.drop_index(columns).run_sync()
        index_names = Concert.indexes().run_sync()
        self.assertNotIn(index_name, index_names)


class TestIndexName(TestCase):
    def test_index_name(self):
        self.assertEqual(
            Manager._get_index_name(["name", "id"]), "manager_name_id"
        )



import time  # For time travel queries.

from piccolo.query.mixins import ColumnsDelegate
from tests.base import DBTestCase, engines_only
from tests.example_apps.music.tables import Band


class TestColumnsDelegate(DBTestCase):
    def test_list_unpacking(self):
        """
        The ``ColumnsDelegate`` should unpack a list of columns if passed in by
        mistake, without the user unpacking them explicitly.

        .. code-block:: python

            # These two should both work the same:
            await Band.select([Band.id, Band.name]).run()
            await Band.select(Band.id, Band.name).run()

        """
        columns_delegate = ColumnsDelegate()

        columns_delegate.columns([Band.name])
        self.assertEqual(columns_delegate.selected_columns, [Band.name])

        columns_delegate.columns([Band.id])
        self.assertEqual(
            columns_delegate.selected_columns, [Band.name, Band.id]
        )

    @engines_only("cockroach")
    def test_as_of(self):
        """
        Time travel queries using "As Of" syntax.
        Currently supports Cockroach using AS OF SYSTEM TIME.
        """
        self.insert_rows()
        time.sleep(1)  # Ensure time travel queries have some history to use!

        result = (
            Band.select()
            .where(Band.name == "Pythonistas")
            .as_of("-500ms")
            .limit(1)
        )
        self.assertTrue("AS OF SYSTEM TIME '-500ms'" in str(result))
        result = result.run_sync()

        self.assertTrue(result[0]["name"] == "Pythonistas")

        result = Band.select().as_of()
        self.assertTrue("AS OF SYSTEM TIME '-1s'" in str(result))
        result = result.run_sync()

        self.assertTrue(result[0]["name"] == "Pythonistas")

        # Alternative syntax.
        result = Band.objects().get(Band.name == "Pythonistas").as_of("-1s")
        self.assertTrue("AS OF SYSTEM TIME '-1s'" in str(result))
        result = result.run_sync()

        self.assertTrue(result.name == "Pythonistas")



from unittest import TestCase

from piccolo.utils.dictionary import make_nested


class TestMakeNested(TestCase):
    def test_nesting(self):
        response = make_nested(
            {
                "id": 1,
                "name": "Pythonistas",
                "manager.id": 1,
                "manager.name": "Guido",
                "manager.car.colour": "green",
            }
        )
        self.assertEqual(
            response,
            {
                "id": 1,
                "name": "Pythonistas",
                "manager": {
                    "id": 1,
                    "name": "Guido",
                    "car": {"colour": "green"},
                },
            },
        )

    def test_name_clash(self):
        """
        In this example, `manager` and `manager.*` could potentially clash.
        Nesting should take precedence.
        """
        response = make_nested(
            {
                "id": 1,
                "name": "Pythonistas",
                "manager": 1,
                "manager.id": 1,
                "manager.name": "Guido",
                "manager.car.colour": "green",
            }
        )
        self.assertEqual(
            response,
            {
                "id": 1,
                "name": "Pythonistas",
                "manager": {
                    "id": 1,
                    "name": "Guido",
                    "car": {"colour": "green"},
                },
            },
        )



from .comparison import (
    ArrayAll,
    ArrayAny,
    Equal,
    GreaterEqualThan,
    GreaterThan,
    ILike,
    In,
    IsNotNull,
    IsNull,
    LessEqualThan,
    LessThan,
    Like,
    NotEqual,
    NotIn,
    NotLike,
)
from .math import Add, Divide, Multiply, Subtract
from .string import Concat



from .alter import Alter
from .count import Count
from .create import Create
from .create_index import CreateIndex
from .delete import Delete
from .drop_index import DropIndex
from .exists import Exists
from .insert import Insert
from .objects import Objects
from .raw import Raw
from .refresh import Refresh
from .select import Avg, Max, Min, Select, Sum
from .table_exists import TableExists
from .update import Update



from unittest import TestCase
from unittest.mock import MagicMock, patch

from piccolo.columns import Secret
from piccolo.columns.column_types import (
    JSON,
    JSONB,
    Email,
    ForeignKey,
    Varchar,
)
from piccolo.table import TABLENAME_WARNING, Table
from tests.example_apps.music.tables import Band


class TestMetaClass(TestCase):
    def test_tablename(self):
        self.assertEqual(Band._meta.tablename, "band")

    def test_protected_table_names(self):
        """
        Some tablenames are forbidden because they're reserved words in the
        database, and can potentially cause issues.
        """
        expected_warning = TABLENAME_WARNING.format(tablename="user")

        with patch("piccolo.table.warnings") as warnings:

            class User(Table):
                pass

            warnings.warn.assert_called_with(expected_warning)

        with patch("piccolo.table.warnings") as warnings:

            class MyUser(Table, tablename="user"):
                pass

            warnings.warn.assert_called_with(expected_warning)

    def test_help_text(self):
        """
        Make sure help_text can be set for the Table.
        """
        help_text = "The manager of a band."

        class Manager(Table, help_text=help_text):
            pass

        self.assertEqual(Manager._meta.help_text, help_text)

    def test_schema(self):
        """
        Make sure schema can be set for the Table.
        """
        schema = "schema_1"

        class Manager(Table, schema=schema):
            pass

        self.assertEqual(Manager._meta.schema, schema)

    @patch("piccolo.table.warnings")
    def test_schema_from_tablename(self, warnings: MagicMock):
        """
        If the tablename contains a '.' we extract the schema name.
        """
        table = "manager"
        schema = "schema_1"

        tablename = f"{schema}.{table}"

        class Manager(Table, tablename=tablename):
            pass

        self.assertEqual(Manager._meta.schema, schema)
        self.assertEqual(Manager._meta.tablename, table)

        warnings.warn.assert_called_once_with(
            "There's a '.' in the tablename - please use the `schema` "
            "argument instead."
        )

    def test_foreign_key_columns(self):
        """
        Make sure TableMeta.foreign_keys and TableMeta.foreign_key_references
        are setup correctly.
        """

        class TableA(Table):
            pass

        class TableB(Table):
            table_a = ForeignKey(references=TableA)

        self.assertEqual(TableB._meta.foreign_key_columns, [TableB.table_a])
        self.assertEqual(TableA._meta.foreign_key_references, [TableB.table_a])

    def test_secret_columns(self):
        """
        Make sure TableMeta.secret_columns are setup correctly.
        """

        class Classified(Table):
            top_secret = Secret()

        self.assertEqual(
            Classified._meta.secret_columns, [Classified.top_secret]
        )

    def test_json_columns(self):
        """
        Make sure TableMeta.json_columns are setup correctly.
        """

        class MyTable(Table):
            column_a = JSON()
            column_b = JSONB()

        self.assertEqual(
            MyTable._meta.json_columns, [MyTable.column_a, MyTable.column_b]
        )

    def test_email_columns(self):
        """
        Make sure ``TableMeta.email_columns`` are setup correctly.
        """

        class MyTable(Table):
            column_a = Email()
            column_b = Varchar()

        self.assertEqual(MyTable._meta.email_columns, [MyTable.column_a])

    def test_id_column(self):
        """
        Makes sure an id column is added.
        """

        class TableA(Table):
            pass

        self.assertTrue(hasattr(TableA, "id"))



from __future__ import annotations

import typing as t
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from piccolo.apps.migrations.auto.schema_differ import (
    DiffableTable,
    RenameColumn,
    RenameColumnCollection,
    RenameTable,
    RenameTableCollection,
    SchemaDiffer,
)
from piccolo.columns.column_types import Numeric, Varchar


class TestSchemaDiffer(TestCase):

    maxDiff = None

    def test_add_table(self):
        """
        Test adding a new table.
        """
        name_column = Varchar()
        name_column._meta.name = "name"
        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band", tablename="band", columns=[name_column]
            )
        ]
        schema_snapshot: t.List[DiffableTable] = []
        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )

        create_tables = schema_differ.create_tables
        self.assertTrue(len(create_tables.statements) == 1)
        self.assertEqual(
            create_tables.statements[0],
            "manager.add_table(class_name='Band', tablename='band', schema=None, columns=None)",  # noqa: E501
        )

        new_table_columns = schema_differ.new_table_columns
        self.assertTrue(len(new_table_columns.statements) == 1)
        self.assertEqual(
            new_table_columns.statements[0],
            "manager.add_column(table_class_name='Band', tablename='band', column_name='name', db_column_name='name', column_class_name='Varchar', column_class=Varchar, params={'length': 255, 'default': '', 'null': False, 'primary_key': False, 'unique': False, 'index': False, 'index_method': IndexMethod.btree, 'choices': None, 'db_column_name': None, 'secret': False})",  # noqa
        )

    def test_drop_table(self):
        """
        Test dropping an existing table.
        """
        schema: t.List[DiffableTable] = []
        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(class_name="Band", tablename="band", columns=[])
        ]
        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )

        self.assertTrue(len(schema_differ.drop_tables.statements) == 1)
        self.assertEqual(
            schema_differ.drop_tables.statements[0],
            "manager.drop_table(class_name='Band', tablename='band', schema=None)",  # noqa: E501
        )

    def test_rename_table(self):
        """
        Test renaming a table.
        """
        name_column = Varchar()
        name_column._meta.name = "name"

        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Act", tablename="act", columns=[name_column]
            )
        ]
        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band", tablename="band", columns=[name_column]
            )
        ]

        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )

        self.assertTrue(len(schema_differ.rename_tables.statements) == 1)
        self.assertEqual(
            schema_differ.rename_tables.statements[0],
            "manager.rename_table(old_class_name='Band', old_tablename='band', new_class_name='Act', new_tablename='act')",  # noqa: E501
        )

        self.assertEqual(schema_differ.create_tables.statements, [])
        self.assertEqual(schema_differ.drop_tables.statements, [])

    def test_change_schema(self):
        """
        Testing changing the schema.
        """
        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[],
                schema="schema_1",
            )
        ]
        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[],
                schema=None,
            )
        ]

        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )

        self.assertEqual(len(schema_differ.change_table_schemas.statements), 1)

        self.assertEqual(
            schema_differ.change_table_schemas.statements[0],
            "manager.change_table_schema(class_name='Band', tablename='band', new_schema='schema_1', old_schema=None)",  # noqa: E501
        )

        self.assertListEqual(schema_differ.create_tables.statements, [])
        self.assertListEqual(schema_differ.drop_tables.statements, [])

    def test_add_column(self):
        """
        Test adding a column to an existing table.
        """
        name_column = Varchar()
        name_column._meta.name = "name"

        genre_column = Varchar()
        genre_column._meta.name = "genre"

        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[name_column, genre_column],
            )
        ]
        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[name_column],
            )
        ]

        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )

        self.assertTrue(len(schema_differ.add_columns.statements) == 1)
        self.assertEqual(
            schema_differ.add_columns.statements[0],
            "manager.add_column(table_class_name='Band', tablename='band', column_name='genre', db_column_name='genre', column_class_name='Varchar', column_class=Varchar, params={'length': 255, 'default': '', 'null': False, 'primary_key': False, 'unique': False, 'index': False, 'index_method': IndexMethod.btree, 'choices': None, 'db_column_name': None, 'secret': False})",  # noqa: E501
        )

    def test_drop_column(self):
        """
        Test dropping a column from an existing table.
        """
        name_column = Varchar()
        name_column._meta.name = "name"

        genre_column = Varchar()
        genre_column._meta.name = "genre"

        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[name_column],
            )
        ]
        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[name_column, genre_column],
            )
        ]

        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )

        self.assertTrue(len(schema_differ.drop_columns.statements) == 1)
        self.assertEqual(
            schema_differ.drop_columns.statements[0],
            "manager.drop_column(table_class_name='Band', tablename='band', column_name='genre', db_column_name='genre')",  # noqa: E501
        )

    def test_rename_column(self):
        """
        Test renaming a column in an existing table.
        """
        # We're going to rename the 'name' column to 'title'
        name_column = Varchar()
        name_column._meta.name = "name"

        title_column = Varchar()
        title_column._meta.name = "title"

        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[title_column],
            )
        ]
        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[name_column],
            )
        ]

        # Test 1 - Tell Piccolo the column was renamed
        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )
        self.assertEqual(schema_differ.add_columns.statements, [])
        self.assertEqual(schema_differ.drop_columns.statements, [])
        self.assertEqual(
            schema_differ.rename_columns.statements,
            [
                "manager.rename_column(table_class_name='Band', tablename='band', old_column_name='title', new_column_name='name', old_db_column_name='title', new_db_column_name='name')"  # noqa: E501
            ],
        )

        # Test 2 - Tell Piccolo the column wasn't renamed
        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="n"
        )
        self.assertEqual(
            schema_differ.add_columns.statements,
            [
                "manager.add_column(table_class_name='Band', tablename='band', column_name='name', db_column_name='name', column_class_name='Varchar', column_class=Varchar, params={'length': 255, 'default': '', 'null': False, 'primary_key': False, 'unique': False, 'index': False, 'index_method': IndexMethod.btree, 'choices': None, 'db_column_name': None, 'secret': False})"  # noqa: E501
            ],
        )
        self.assertEqual(
            schema_differ.drop_columns.statements,
            [
                "manager.drop_column(table_class_name='Band', tablename='band', column_name='title', db_column_name='title')"  # noqa: E501
            ],
        )
        self.assertTrue(schema_differ.rename_columns.statements == [])

    @patch("piccolo.apps.migrations.auto.schema_differ.input")
    def test_rename_multiple_columns(self, input: MagicMock):
        """
        Make sure renaming columns works when several columns have been
        renamed.
        """
        # We're going to rename a1 to a2, and b1 to b2.
        a1 = Varchar()
        a1._meta.name = "a1"

        a2 = Varchar()
        a2._meta.name = "a2"

        b1 = Varchar()
        b1._meta.name = "b1"

        b2 = Varchar()
        b2._meta.name = "b2"

        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[a1, b1],
            )
        ]
        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[a2, b2],
            )
        ]

        def mock_input(value: str):
            """
            We need to dynamically set the return value based on what's passed
            in.
            """
            return (
                "y"
                if value
                in (
                    "Did you rename the `a1` column to `a2` on the `Band` table? (y/N)",  # noqa: E501
                    "Did you rename the `b1` column to `b2` on the `Band` table? (y/N)",  # noqa: E501
                )
                else "n"
            )

        input.side_effect = mock_input

        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot
        )
        self.assertEqual(schema_differ.add_columns.statements, [])
        self.assertEqual(schema_differ.drop_columns.statements, [])
        self.assertEqual(
            schema_differ.rename_columns.statements,
            [
                "manager.rename_column(table_class_name='Band', tablename='band', old_column_name='a1', new_column_name='a2', old_db_column_name='a1', new_db_column_name='a2')",  # noqa: E501
                "manager.rename_column(table_class_name='Band', tablename='band', old_column_name='b1', new_column_name='b2', old_db_column_name='b1', new_db_column_name='b2')",  # noqa: E501
            ],
        )

        self.assertEqual(
            input.call_args_list,
            [
                call(
                    "Did you rename the `a1` column to `a2` on the `Band` table? (y/N)"  # noqa: E501
                ),
                call(
                    "Did you rename the `b1` column to `b2` on the `Band` table? (y/N)"  # noqa: E501
                ),
            ],
        )

    @patch("piccolo.apps.migrations.auto.schema_differ.input")
    def test_rename_some_columns(self, input: MagicMock):
        """
        Make sure that some columns can be marked as renamed, and others are
        dropped / created.
        """
        # We're going to rename a1 to a2, but want b1 to be dropped, and b2 to
        # be created.
        a1 = Varchar()
        a1._meta.name = "a1"

        a2 = Varchar()
        a2._meta.name = "a2"

        b1 = Varchar()
        b1._meta.name = "b1"

        b2 = Varchar()
        b2._meta.name = "b2"

        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[a1, b1],
            )
        ]
        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Band",
                tablename="band",
                columns=[a2, b2],
            )
        ]

        def mock_input(value: str):
            """
            We need to dynamically set the return value based on what's passed
            in.
            """
            return (
                "y"
                if value
                == "Did you rename the `a1` column to `a2` on the `Band` table? (y/N)"  # noqa: E501
                else "n"
            )

        input.side_effect = mock_input

        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot
        )
        self.assertEqual(
            schema_differ.add_columns.statements,
            [
                "manager.add_column(table_class_name='Band', tablename='band', column_name='b2', db_column_name='b2', column_class_name='Varchar', column_class=Varchar, params={'length': 255, 'default': '', 'null': False, 'primary_key': False, 'unique': False, 'index': False, 'index_method': IndexMethod.btree, 'choices': None, 'db_column_name': None, 'secret': False})"  # noqa: E501
            ],
        )
        self.assertEqual(
            schema_differ.drop_columns.statements,
            [
                "manager.drop_column(table_class_name='Band', tablename='band', column_name='b1', db_column_name='b1')"  # noqa: E501
            ],
        )
        self.assertEqual(
            schema_differ.rename_columns.statements,
            [
                "manager.rename_column(table_class_name='Band', tablename='band', old_column_name='a1', new_column_name='a2', old_db_column_name='a1', new_db_column_name='a2')",  # noqa: E501
            ],
        )

        self.assertEqual(
            input.call_args_list,
            [
                call(
                    "Did you rename the `a1` column to `a2` on the `Band` table? (y/N)"  # noqa: E501
                ),
                call(
                    "Did you rename the `b1` column to `b2` on the `Band` table? (y/N)"  # noqa: E501
                ),
            ],
        )

    def test_alter_column_precision(self):
        price_1 = Numeric(digits=(4, 2))
        price_1._meta.name = "price"

        price_2 = Numeric(digits=(5, 2))
        price_2._meta.name = "price"

        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Ticket",
                tablename="ticket",
                columns=[price_1],
            )
        ]
        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Ticket",
                tablename="ticket",
                columns=[price_2],
            )
        ]

        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )

        self.assertTrue(len(schema_differ.alter_columns.statements) == 1)
        self.assertEqual(
            schema_differ.alter_columns.statements[0],
            "manager.alter_column(table_class_name='Ticket', tablename='ticket', column_name='price', db_column_name='price', params={'digits': (4, 2)}, old_params={'digits': (5, 2)}, column_class=Numeric, old_column_class=Numeric)",  # noqa
        )

    def test_db_column_name(self):
        """
        Make sure alter statements use the ``db_column_name`` if provided.

        https://github.com/piccolo-orm/piccolo/issues/513

        """
        price_1 = Numeric(digits=(4, 2), db_column_name="custom")
        price_1._meta.name = "price"

        price_2 = Numeric(digits=(5, 2), db_column_name="custom")
        price_2._meta.name = "price"

        schema: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Ticket",
                tablename="ticket",
                columns=[price_1],
            )
        ]
        schema_snapshot: t.List[DiffableTable] = [
            DiffableTable(
                class_name="Ticket",
                tablename="ticket",
                columns=[price_2],
            )
        ]

        schema_differ = SchemaDiffer(
            schema=schema, schema_snapshot=schema_snapshot, auto_input="y"
        )

        self.assertTrue(len(schema_differ.alter_columns.statements) == 1)
        self.assertEqual(
            schema_differ.alter_columns.statements[0],
            "manager.alter_column(table_class_name='Ticket', tablename='ticket', column_name='price', db_column_name='custom', params={'digits': (4, 2)}, old_params={'digits': (5, 2)}, column_class=Numeric, old_column_class=Numeric)",  # noqa
        )

    def test_alter_default(self):
        pass


class TestRenameTableCollection(TestCase):
    collection = RenameTableCollection(
        rename_tables=[
            RenameTable(
                old_class_name="Manager",
                old_tablename="manager",
                new_class_name="Manager1",
                new_tablename="manager_1",
            )
        ]
    )

    def test_was_renamed_from(self):
        self.assertTrue(
            self.collection.was_renamed_from(old_class_name="Manager")
        )
        self.assertFalse(
            self.collection.was_renamed_from(old_class_name="Band")
        )

    def test_renamed_from(self):
        self.assertEqual(
            self.collection.renamed_from(new_class_name="Manager1"), "Manager"
        )
        self.assertIsNone(
            self.collection.renamed_from(new_class_name="Band"),
        )


class TestRenameColumnCollection(TestCase):
    def test_for_table_class_name(self):
        rename_column = RenameColumn(
            table_class_name="Manager",
            tablename="manager",
            old_column_name="name",
            new_column_name="full_name",
            old_db_column_name="name",
            new_db_column_name="full_name",
        )

        collection = RenameColumnCollection(rename_columns=[rename_column])

        self.assertListEqual(
            collection.for_table_class_name(table_class_name="Manager"),
            [rename_column],
        )
        self.assertListEqual(
            collection.for_table_class_name(table_class_name="Band"), []
        )



import typing as t

from piccolo.engine.finder import engine_finder
from piccolo.table import Table, create_db_tables_sync, drop_db_tables_sync
from tests.base import engine_is, engines_skip

engine = engine_finder()


class M2MBase:
    """
    This allows us to test M2M when the tables are in different schemas
    (public vs non-public).
    """

    band: t.Type[Table]
    genre: t.Type[Table]
    genre_to_band: t.Type[Table]
    all_tables: t.List[t.Type[Table]]

    def setUp(self):
        Band = self.band
        Genre = self.genre
        GenreToBand = self.genre_to_band

        create_db_tables_sync(*self.all_tables, if_not_exists=True)

        if engine_is("cockroach"):
            bands = (
                Band.insert(
                    Band(name="Pythonistas"),
                    Band(name="Rustaceans"),
                    Band(name="C-Sharps"),
                )
                .returning(Band.id)
                .run_sync()
            )

            genres = (
                Genre.insert(
                    Genre(name="Rock"),
                    Genre(name="Folk"),
                    Genre(name="Classical"),
                )
                .returning(Genre.id)
                .run_sync()
            )

            GenreToBand.insert(
                GenreToBand(band=bands[0]["id"], genre=genres[0]["id"]),
                GenreToBand(band=bands[0]["id"], genre=genres[1]["id"]),
                GenreToBand(band=bands[1]["id"], genre=genres[1]["id"]),
                GenreToBand(band=bands[2]["id"], genre=genres[0]["id"]),
                GenreToBand(band=bands[2]["id"], genre=genres[2]["id"]),
            ).run_sync()
        else:
            Band.insert(
                Band(name="Pythonistas"),
                Band(name="Rustaceans"),
                Band(name="C-Sharps"),
            ).run_sync()

            Genre.insert(
                Genre(name="Rock"),
                Genre(name="Folk"),
                Genre(name="Classical"),
            ).run_sync()

            GenreToBand.insert(
                GenreToBand(band=1, genre=1),
                GenreToBand(band=1, genre=2),
                GenreToBand(band=2, genre=2),
                GenreToBand(band=3, genre=1),
                GenreToBand(band=3, genre=3),
            ).run_sync()

    def tearDown(self):
        drop_db_tables_sync(*self.all_tables)

    @engines_skip("cockroach")
    def test_select_name(self):
        """
        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg
        """  # noqa: E501
        Band = self.band
        Genre = self.genre

        response = Band.select(
            Band.name, Band.genres(Genre.name, as_list=True)
        ).run_sync()
        self.assertEqual(
            response,
            [
                {"name": "Pythonistas", "genres": ["Rock", "Folk"]},
                {"name": "Rustaceans", "genres": ["Folk"]},
                {"name": "C-Sharps", "genres": ["Rock", "Classical"]},
            ],
        )

        # Now try it in reverse.
        response = Genre.select(
            Genre.name, Genre.bands(Band.name, as_list=True)
        ).run_sync()
        self.assertEqual(
            response,
            [
                {"name": "Rock", "bands": ["Pythonistas", "C-Sharps"]},
                {"name": "Folk", "bands": ["Pythonistas", "Rustaceans"]},
                {"name": "Classical", "bands": ["C-Sharps"]},
            ],
        )

    @engines_skip("cockroach")
    def test_no_related(self):
        """
        Make sure it still works correctly if there are no related values.
        """
        """
        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg
        """  # noqa: E501
        Band = self.band
        Genre = self.genre
        GenreToBand = self.genre_to_band

        GenreToBand.delete(force=True).run_sync()

        # Try it with a list response
        response = Band.select(
            Band.name, Band.genres(Genre.name, as_list=True)
        ).run_sync()

        self.assertEqual(
            response,
            [
                {"name": "Pythonistas", "genres": []},
                {"name": "Rustaceans", "genres": []},
                {"name": "C-Sharps", "genres": []},
            ],
        )

        # Also try it with a nested response
        response = Band.select(
            Band.name, Band.genres(Genre.id, Genre.name)
        ).run_sync()
        self.assertEqual(
            response,
            [
                {"name": "Pythonistas", "genres": []},
                {"name": "Rustaceans", "genres": []},
                {"name": "C-Sharps", "genres": []},
            ],
        )

    @engines_skip("cockroach")
    def test_select_multiple(self):
        """
        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg
        """  # noqa: E501
        Band = self.band
        Genre = self.genre

        response = Band.select(
            Band.name, Band.genres(Genre.id, Genre.name)
        ).run_sync()

        self.assertEqual(
            response,
            [
                {
                    "name": "Pythonistas",
                    "genres": [
                        {"id": 1, "name": "Rock"},
                        {"id": 2, "name": "Folk"},
                    ],
                },
                {"name": "Rustaceans", "genres": [{"id": 2, "name": "Folk"}]},
                {
                    "name": "C-Sharps",
                    "genres": [
                        {"id": 1, "name": "Rock"},
                        {"id": 3, "name": "Classical"},
                    ],
                },
            ],
        )

        # Now try it in reverse.
        response = Genre.select(
            Genre.name, Genre.bands(Band.id, Band.name)
        ).run_sync()

        self.assertEqual(
            response,
            [
                {
                    "name": "Rock",
                    "bands": [
                        {"id": 1, "name": "Pythonistas"},
                        {"id": 3, "name": "C-Sharps"},
                    ],
                },
                {
                    "name": "Folk",
                    "bands": [
                        {"id": 1, "name": "Pythonistas"},
                        {"id": 2, "name": "Rustaceans"},
                    ],
                },
                {
                    "name": "Classical",
                    "bands": [{"id": 3, "name": "C-Sharps"}],
                },
            ],
        )

    @engines_skip("cockroach")
    def test_select_id(self):
        """
        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg
        """  # noqa: E501
        Band = self.band
        Genre = self.genre

        response = Band.select(
            Band.name, Band.genres(Genre.id, as_list=True)
        ).run_sync()
        self.assertEqual(
            response,
            [
                {"name": "Pythonistas", "genres": [1, 2]},
                {"name": "Rustaceans", "genres": [2]},
                {"name": "C-Sharps", "genres": [1, 3]},
            ],
        )

        # Now try it in reverse.
        response = Genre.select(
            Genre.name, Genre.bands(Band.id, as_list=True)
        ).run_sync()
        self.assertEqual(
            response,
            [
                {"name": "Rock", "bands": [1, 3]},
                {"name": "Folk", "bands": [1, 2]},
                {"name": "Classical", "bands": [3]},
            ],
        )

    @engines_skip("cockroach")
    def test_select_all_columns(self):
        """
        Make sure ``all_columns`` can be passed in as an argument. ``M2M``
        should flatten the arguments. Reported here:

        https://github.com/piccolo-orm/piccolo/issues/728

        🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg

        """  # noqa: E501
        Band = self.band
        Genre = self.genre

        response = Band.select(
            Band.name, Band.genres(Genre.all_columns(exclude=(Genre.id,)))
        ).run_sync()
        self.assertEqual(
            response,
            [
                {
                    "name": "Pythonistas",
                    "genres": [
                        {"name": "Rock"},
                        {"name": "Folk"},
                    ],
                },
                {"name": "Rustaceans", "genres": [{"name": "Folk"}]},
                {
                    "name": "C-Sharps",
                    "genres": [
                        {"name": "Rock"},
                        {"name": "Classical"},
                    ],
                },
            ],
        )

    def test_add_m2m(self):
        """
        Make sure we can add items to the joining table.
        """
        Band = self.band
        Genre = self.genre
        GenreToBand = self.genre_to_band

        band: Band = Band.objects().get(Band.name == "Pythonistas").run_sync()
        band.add_m2m(Genre(name="Punk Rock"), m2m=Band.genres).run_sync()

        self.assertTrue(
            Genre.exists().where(Genre.name == "Punk Rock").run_sync()
        )

        self.assertEqual(
            GenreToBand.count()
            .where(
                GenreToBand.band.name == "Pythonistas",
                GenreToBand.genre.name == "Punk Rock",
            )
            .run_sync(),
            1,
        )

    def test_extra_columns_str(self):
        """
        Make sure the ``extra_column_values`` parameter for ``add_m2m`` works
        correctly when the dictionary keys are strings.
        """
        Band = self.band
        Genre = self.genre
        GenreToBand = self.genre_to_band

        reason = "Their second album was very punk rock."

        band: Band = Band.objects().get(Band.name == "Pythonistas").run_sync()
        band.add_m2m(
            Genre(name="Punk Rock"),
            m2m=Band.genres,
            extra_column_values={
                "reason": "Their second album was very punk rock."
            },
        ).run_sync()

        genre_to_band = (
            GenreToBand.objects()
            .get(
                (GenreToBand.band.name == "Pythonistas")
                & (GenreToBand.genre.name == "Punk Rock")
            )
            .run_sync()
        )

        self.assertEqual(genre_to_band.reason, reason)

    def test_extra_columns_class(self):
        """
        Make sure the ``extra_column_values`` parameter for ``add_m2m`` works
        correctly when the dictionary keys are ``Column`` classes.
        """
        Band = self.band
        Genre = self.genre
        GenreToBand = self.genre_to_band

        reason = "Their second album was very punk rock."

        band: Band = Band.objects().get(Band.name == "Pythonistas").run_sync()
        band.add_m2m(
            Genre(name="Punk Rock"),
            m2m=Band.genres,
            extra_column_values={
                GenreToBand.reason: "Their second album was very punk rock."
            },
        ).run_sync()

        genre_to_band = (
            GenreToBand.objects()
            .get(
                (GenreToBand.band.name == "Pythonistas")
                & (GenreToBand.genre.name == "Punk Rock")
            )
            .run_sync()
        )

        self.assertEqual(genre_to_band.reason, reason)

    def test_add_m2m_existing(self):
        """
        Make sure we can add an existing element to the joining table.
        """
        Band = self.band
        Genre = self.genre
        GenreToBand = self.genre_to_band

        band: Band = Band.objects().get(Band.name == "Pythonistas").run_sync()

        genre: Genre = (
            Genre.objects().get(Genre.name == "Classical").run_sync()
        )

        band.add_m2m(genre, m2m=Band.genres).run_sync()

        # We shouldn't have created a duplicate genre in the database.
        self.assertEqual(
            Genre.count().where(Genre.name == "Classical").run_sync(), 1
        )

        self.assertEqual(
            GenreToBand.count()
            .where(
                GenreToBand.band.name == "Pythonistas",
                GenreToBand.genre.name == "Classical",
            )
            .run_sync(),
            1,
        )

    def test_get_m2m(self):
        """
        Make sure we can get related items via the joining table.
        """
        Band = self.band

        band: Band = Band.objects().get(Band.name == "Pythonistas").run_sync()

        genres = band.get_m2m(Band.genres).run_sync()

        self.assertTrue(all(isinstance(i, Table) for i in genres))

        self.assertEqual([i.name for i in genres], ["Rock", "Folk"])

    def test_remove_m2m(self):
        """
        Make sure we can remove related items via the joining table.
        """
        Band = self.band
        Genre = self.genre
        GenreToBand = self.genre_to_band

        band: Band = Band.objects().get(Band.name == "Pythonistas").run_sync()

        genre = Genre.objects().get(Genre.name == "Rock").run_sync()

        band.remove_m2m(genre, m2m=Band.genres).run_sync()

        self.assertEqual(
            GenreToBand.count()
            .where(
                GenreToBand.band.name == "Pythonistas",
                GenreToBand.genre.name == "Rock",
            )
            .run_sync(),
            0,
        )

        # Make sure the others weren't removed:
        self.assertEqual(
            GenreToBand.count()
            .where(
                GenreToBand.band.name == "Pythonistas",
                GenreToBand.genre.name == "Folk",
            )
            .run_sync(),
            1,
        )

        self.assertEqual(
            GenreToBand.count()
            .where(
                GenreToBand.band.name == "C-Sharps",
                GenreToBand.genre.name == "Rock",
            )
            .run_sync(),
            1,
        )



import dataclasses
import datetime
import typing as t
from unittest import TestCase

import pytest

from piccolo.columns.base import Column
from piccolo.columns.column_types import (
    Date,
    Integer,
    Interval,
    Text,
    Timestamp,
    Timestamptz,
    Varchar,
)
from piccolo.querystring import QueryString
from piccolo.table import Table
from tests.base import (
    DBTestCase,
    engine_version_lt,
    engines_skip,
    is_running_sqlite,
    sqlite_only,
)
from tests.example_apps.music.tables import Band, Manager


class TestUpdate(DBTestCase):
    def check_response(self):
        response = (
            Band.select(Band.name)
            .where(Band.name == "Pythonistas3")
            .run_sync()
        )

        self.assertEqual(response, [{"name": "Pythonistas3"}])

    def test_update(self):
        """
        Make sure updating work, when passing the new values directly to the
        `update` method.
        """
        self.insert_rows()

        Band.update({Band.name: "Pythonistas3"}).where(
            Band.name == "Pythonistas"
        ).run_sync()

        self.check_response()

    def test_update_with_string_keys(self):
        """
        Make sure updating work, when passing a dictionary of values, which
        uses column names as keys, instead of Column instances.
        """
        self.insert_rows()

        Band.update({"name": "Pythonistas3"}).where(
            Band.name == "Pythonistas"
        ).run_sync()

        self.check_response()

    def test_update_with_kwargs(self):
        """
        Make sure updating work, when passing the new value via kwargs.
        """
        self.insert_rows()

        Band.update(name="Pythonistas3").where(
            Band.name == "Pythonistas"
        ).run_sync()

        self.check_response()

    def test_update_values(self):
        """
        Make sure updating work, when passing the new values via the `values`
        method.
        """
        self.insert_rows()

        Band.update().values({Band.name: "Pythonistas3"}).where(
            Band.name == "Pythonistas"
        ).run_sync()

        self.check_response()

    def test_update_values_with_string_keys(self):
        """
        Make sure updating work, when passing the new values via the `values`
        method, using a column name as a dictionary key.
        """
        self.insert_rows()

        Band.update().values({"name": "Pythonistas3"}).where(
            Band.name == "Pythonistas"
        ).run_sync()

        self.check_response()

    def test_update_values_with_kwargs(self):
        """
        Make sure updating work, when passing the new values via kwargs.
        """
        self.insert_rows()

        Band.update().values(name="Pythonistas3").where(
            Band.name == "Pythonistas"
        ).run_sync()

        self.check_response()

    @pytest.mark.skipif(
        is_running_sqlite() and engine_version_lt(3.35),
        reason="SQLite version not supported",
    )
    def test_update_returning(self):
        """
        Make sure update works with the `returning` clause.
        """
        self.insert_rows()

        response = (
            Band.update({Band.name: "Pythonistas 2"})
            .where(Band.name == "Pythonistas")
            .returning(Band.name)
            .run_sync()
        )

        self.assertEqual(response, [{"name": "Pythonistas 2"}])

    @pytest.mark.skipif(
        is_running_sqlite() and engine_version_lt(3.35),
        reason="SQLite version not supported",
    )
    def test_update_returning_alias(self):
        """
        Make sure update works with the `returning` clause.
        """
        self.insert_rows()

        response = (
            Band.update({Band.name: "Pythonistas 2"})
            .where(Band.name == "Pythonistas")
            .returning(Band.name.as_alias("band name"))
            .run_sync()
        )

        self.assertEqual(response, [{"band name": "Pythonistas 2"}])


###############################################################################
# Test operators


class MyTable(Table):
    integer = Integer(null=True)
    timestamp = Timestamp(null=True)
    timestamptz = Timestamptz(null=True)
    date = Date(null=True)
    interval = Interval(null=True)
    varchar = Varchar(null=True)
    text = Text(null=True)


INITIAL_DATETIME = datetime.datetime(
    year=2022, month=1, day=1, hour=21, minute=0
)
INITIAL_INTERVAL = datetime.timedelta(days=1, hours=1, minutes=1)

DATETIME_DELTA = datetime.timedelta(
    days=1, hours=1, minutes=1, seconds=30, microseconds=1000
)
DATE_DELTA = datetime.timedelta(days=1)


@dataclasses.dataclass
class OperatorTestCase:
    description: str
    column: Column
    initial: t.Any
    querystring: QueryString
    expected: t.Any


TEST_CASES = [
    # Text
    OperatorTestCase(
        description="Add Text",
        column=MyTable.text,
        initial="Pythonistas",
        querystring=MyTable.text + "!!!",
        expected="Pythonistas!!!",
    ),
    OperatorTestCase(
        description="Add Text columns",
        column=MyTable.text,
        initial="Pythonistas",
        querystring=MyTable.text + MyTable.text,
        expected="PythonistasPythonistas",
    ),
    OperatorTestCase(
        description="Reverse add Text",
        column=MyTable.text,
        initial="Pythonistas",
        querystring="!!!" + MyTable.text,
        expected="!!!Pythonistas",
    ),
    OperatorTestCase(
        description="Text is null",
        column=MyTable.text,
        initial=None,
        querystring=MyTable.text + "!!!",
        expected=None,
    ),
    OperatorTestCase(
        description="Reverse Text is null",
        column=MyTable.text,
        initial=None,
        querystring="!!!" + MyTable.text,
        expected=None,
    ),
    # Varchar
    OperatorTestCase(
        description="Add Varchar",
        column=MyTable.varchar,
        initial="Pythonistas",
        querystring=MyTable.varchar + "!!!",
        expected="Pythonistas!!!",
    ),
    OperatorTestCase(
        description="Add Varchar columns",
        column=MyTable.varchar,
        initial="Pythonistas",
        querystring=MyTable.varchar + MyTable.varchar,
        expected="PythonistasPythonistas",
    ),
    OperatorTestCase(
        description="Reverse add Varchar",
        column=MyTable.varchar,
        initial="Pythonistas",
        querystring="!!!" + MyTable.varchar,
        expected="!!!Pythonistas",
    ),
    OperatorTestCase(
        description="Varchar is null",
        column=MyTable.varchar,
        initial=None,
        querystring=MyTable.varchar + "!!!",
        expected=None,
    ),
    OperatorTestCase(
        description="Reverse Varchar is null",
        column=MyTable.varchar,
        initial=None,
        querystring="!!!" + MyTable.varchar,
        expected=None,
    ),
    # Integer
    OperatorTestCase(
        description="Add Integer",
        column=MyTable.integer,
        initial=1000,
        querystring=MyTable.integer + 10,
        expected=1010,
    ),
    OperatorTestCase(
        description="Reverse add Integer",
        column=MyTable.integer,
        initial=1000,
        querystring=10 + MyTable.integer,
        expected=1010,
    ),
    OperatorTestCase(
        description="Add Integer colums together",
        column=MyTable.integer,
        initial=1000,
        querystring=MyTable.integer + MyTable.integer,
        expected=2000,
    ),
    OperatorTestCase(
        description="Subtract Integer",
        column=MyTable.integer,
        initial=1000,
        querystring=MyTable.integer - 10,
        expected=990,
    ),
    OperatorTestCase(
        description="Reverse subtract Integer",
        column=MyTable.integer,
        initial=1000,
        querystring=2000 - MyTable.integer,
        expected=1000,
    ),
    OperatorTestCase(
        description="Multiply Integer",
        column=MyTable.integer,
        initial=1000,
        querystring=MyTable.integer * 2,
        expected=2000,
    ),
    OperatorTestCase(
        description="Reverse multiply Integer",
        column=MyTable.integer,
        initial=1000,
        querystring=2 * MyTable.integer,
        expected=2000,
    ),
    OperatorTestCase(
        description="Divide Integer",
        column=MyTable.integer,
        initial=1000,
        querystring=MyTable.integer / 10,
        expected=100,
    ),
    OperatorTestCase(
        description="Reverse divide Integer",
        column=MyTable.integer,
        initial=1000,
        querystring=2000 / MyTable.integer,
        expected=2,
    ),
    OperatorTestCase(
        description="Integer is null",
        column=MyTable.integer,
        initial=None,
        querystring=MyTable.integer + 1,
        expected=None,
    ),
    OperatorTestCase(
        description="Reverse Integer is null",
        column=MyTable.integer,
        initial=None,
        querystring=1 + MyTable.integer,
        expected=None,
    ),
    # Timestamp
    OperatorTestCase(
        description="Add Timestamp",
        column=MyTable.timestamp,
        initial=INITIAL_DATETIME,
        querystring=MyTable.timestamp + DATETIME_DELTA,
        expected=datetime.datetime(
            year=2022,
            month=1,
            day=2,
            hour=22,
            minute=1,
            second=30,
            microsecond=1000,
        ),
    ),
    OperatorTestCase(
        description="Reverse add Timestamp",
        column=MyTable.timestamp,
        initial=INITIAL_DATETIME,
        querystring=DATETIME_DELTA + MyTable.timestamp,
        expected=datetime.datetime(
            year=2022,
            month=1,
            day=2,
            hour=22,
            minute=1,
            second=30,
            microsecond=1000,
        ),
    ),
    OperatorTestCase(
        description="Subtract Timestamp",
        column=MyTable.timestamp,
        initial=INITIAL_DATETIME,
        querystring=MyTable.timestamp - DATETIME_DELTA,
        expected=datetime.datetime(
            year=2021,
            month=12,
            day=31,
            hour=19,
            minute=58,
            second=29,
            microsecond=999000,
        ),
    ),
    OperatorTestCase(
        description="Timestamp is null",
        column=MyTable.timestamp,
        initial=None,
        querystring=MyTable.timestamp + DATETIME_DELTA,
        expected=None,
    ),
    # Timestamptz
    OperatorTestCase(
        description="Add Timestamptz",
        column=MyTable.timestamptz,
        initial=INITIAL_DATETIME,
        querystring=MyTable.timestamptz + DATETIME_DELTA,
        expected=datetime.datetime(
            year=2022,
            month=1,
            day=2,
            hour=22,
            minute=1,
            second=30,
            microsecond=1000,
            tzinfo=datetime.timezone.utc,
        ),
    ),
    OperatorTestCase(
        description="Reverse add Timestamptz",
        column=MyTable.timestamptz,
        initial=INITIAL_DATETIME,
        querystring=DATETIME_DELTA + MyTable.timestamptz,
        expected=datetime.datetime(
            year=2022,
            month=1,
            day=2,
            hour=22,
            minute=1,
            second=30,
            microsecond=1000,
            tzinfo=datetime.timezone.utc,
        ),
    ),
    OperatorTestCase(
        description="Subtract Timestamptz",
        column=MyTable.timestamptz,
        initial=INITIAL_DATETIME,
        querystring=MyTable.timestamptz - DATETIME_DELTA,
        expected=datetime.datetime(
            year=2021,
            month=12,
            day=31,
            hour=19,
            minute=58,
            second=29,
            microsecond=999000,
            tzinfo=datetime.timezone.utc,
        ),
    ),
    OperatorTestCase(
        description="Timestamptz is null",
        column=MyTable.timestamptz,
        initial=None,
        querystring=MyTable.timestamptz + DATETIME_DELTA,
        expected=None,
    ),
    # Date
    OperatorTestCase(
        description="Add Date",
        column=MyTable.date,
        initial=INITIAL_DATETIME,
        querystring=MyTable.date + DATE_DELTA,
        expected=datetime.date(year=2022, month=1, day=2),
    ),
    OperatorTestCase(
        description="Reverse add Date",
        column=MyTable.date,
        initial=INITIAL_DATETIME,
        querystring=DATE_DELTA + MyTable.date,
        expected=datetime.date(year=2022, month=1, day=2),
    ),
    OperatorTestCase(
        description="Subtract Date",
        column=MyTable.date,
        initial=INITIAL_DATETIME,
        querystring=MyTable.date - DATE_DELTA,
        expected=datetime.date(year=2021, month=12, day=31),
    ),
    OperatorTestCase(
        description="Date is null",
        column=MyTable.date,
        initial=None,
        querystring=MyTable.date + DATE_DELTA,
        expected=None,
    ),
    # Interval
    OperatorTestCase(
        description="Add Interval",
        column=MyTable.interval,
        initial=INITIAL_INTERVAL,
        querystring=MyTable.interval + DATETIME_DELTA,
        expected=datetime.timedelta(days=2, seconds=7350, microseconds=1000),
    ),
    OperatorTestCase(
        description="Reverse add Interval",
        column=MyTable.interval,
        initial=INITIAL_INTERVAL,
        querystring=DATETIME_DELTA + MyTable.interval,
        expected=datetime.timedelta(days=2, seconds=7350, microseconds=1000),
    ),
    OperatorTestCase(
        description="Subtract Interval",
        column=MyTable.interval,
        initial=INITIAL_INTERVAL,
        querystring=MyTable.interval - DATETIME_DELTA,
        expected=datetime.timedelta(
            days=-1, seconds=86369, microseconds=999000
        ),
    ),
    OperatorTestCase(
        description="Interval is null",
        column=MyTable.interval,
        initial=None,
        querystring=MyTable.interval + DATETIME_DELTA,
        expected=None,
    ),
]


class TestOperators(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    @engines_skip("cockroach")
    def test_operators(self):
        for test_case in TEST_CASES:
            print(test_case.description)

            # Create the initial data in the database.
            instance = MyTable()
            setattr(instance, test_case.column._meta.name, test_case.initial)
            instance.save().run_sync()

            # Apply the update.
            MyTable.update(
                {test_case.column: test_case.querystring}, force=True
            ).run_sync()

            # Make sure the value returned from the database is correct.
            new_value = getattr(
                MyTable.objects().first().run_sync(),
                test_case.column._meta.name,
            )

            self.assertEqual(
                new_value, test_case.expected, msg=test_case.description
            )

            # Clean up
            MyTable.delete(force=True).run_sync()

    @sqlite_only
    def test_edge_cases(self):
        """
        Some usecases aren't supported by SQLite, and should raise a
        ``ValueError``.
        """
        with self.assertRaises(ValueError):
            # An error should be raised because we can't save at this level
            # of resolution - 1 millisecond is the minimum.
            MyTable.timestamp + datetime.timedelta(microseconds=1)


###############################################################################
# Test auto_update


class AutoUpdateTable(Table, tablename="my_table"):
    name = Varchar()
    modified_on = Timestamp(
        auto_update=datetime.datetime.now, null=True, default=None
    )


class TestAutoUpdate(TestCase):
    def setUp(self):
        AutoUpdateTable.create_table().run_sync()

    def tearDown(self):
        AutoUpdateTable.alter().drop_table().run_sync()

    def test_save(self):
        """
        Make sure the ``save`` method uses ``auto_update`` columns correctly.
        """
        row = AutoUpdateTable(name="test")

        # Saving for the first time is an INSERT, so `auto_update` shouldn't
        # be triggered.
        row.save().run_sync()
        self.assertIsNone(row.modified_on)

        # A subsequent save is an UPDATE, so `auto_update` should be triggered.
        row.name = "test 2"
        row.save().run_sync()
        self.assertIsInstance(row.modified_on, datetime.datetime)

        # If we save it again, `auto_update` should be applied again.
        existing_modified_on = row.modified_on
        row.name = "test 3"
        row.save().run_sync()
        self.assertIsInstance(row.modified_on, datetime.datetime)
        self.assertGreater(row.modified_on, existing_modified_on)

    def test_update(self):
        """
        Make sure the update method uses ``auto_update`` columns correctly.
        """
        # Insert a row for us to update
        AutoUpdateTable.insert(AutoUpdateTable(name="test")).run_sync()

        self.assertDictEqual(
            AutoUpdateTable.select(
                AutoUpdateTable.name, AutoUpdateTable.modified_on
            )
            .first()
            .run_sync(),
            {"name": "test", "modified_on": None},
        )

        # Update the row
        AutoUpdateTable.update(
            {AutoUpdateTable.name: "test 2"}, force=True
        ).run_sync()

        # Retrieve the row
        updated_row = (
            AutoUpdateTable.select(
                AutoUpdateTable.name, AutoUpdateTable.modified_on
            )
            .first()
            .run_sync()
        )
        self.assertIsInstance(updated_row["modified_on"], datetime.datetime)
        self.assertEqual(updated_row["name"], "test 2")


###############################################################################
# Test update with joins


class TestUpdateWithJoin(DBTestCase):
    def test_join(self):
        """
        Make sure updates work when the where clause needs a join.
        """
        self.insert_rows()
        Band.update({Band.name: "New name"}).where(
            Band.manager.name == "Guido"
        ).run_sync()

        self.assertEqual(
            Band.count().where(Band.name == "New name").run_sync(), 1
        )

    def test_multiple_matches(self):
        """
        Make sure it works when the join has multiple matching values.
        """
        self.insert_rows()

        # Create an additional band with the same manager.
        manager = Manager.objects().get(Manager.name == "Guido").run_sync()
        band = Band(name="Pythonistas 2", manager=manager)
        band.save().run_sync()

        Band.update({Band.name: "New name"}).where(
            Band.manager.name == "Guido"
        ).run_sync()

        self.assertEqual(
            Band.count().where(Band.name == "New name").run_sync(), 2
        )

    def test_no_matches(self):
        """
        Make sure it works when the join has no matching values.
        """
        self.insert_rows()

        Band.update({Band.name: "New name"}).where(
            Band.manager.name == "Mr Manager"
        ).run_sync()

        self.assertEqual(
            Band.count().where(Band.name == "New name").run_sync(), 0
        )

    def test_and(self):
        """
        Make sure it works when combined with other where clauses using AND.
        """
        self.insert_rows()

        # Create an additional band with the same manager, and different
        # popularity.
        manager = Manager.objects().get(Manager.name == "Guido").run_sync()
        band = Band(name="Pythonistas 2", manager=manager, popularity=10000)
        band.save().run_sync()

        Band.update({Band.name: "New name"}).where(
            Band.manager.name == "Guido", Band.popularity == 10000
        ).run_sync()

        self.assertEqual(
            Band.count().where(Band.name == "New name").run_sync(), 1
        )

    def test_or(self):
        """
        Make sure it works when combined with other where clauses using OR.
        """
        self.insert_rows()

        Band.update({Band.name: "New name"}).where(
            (Band.manager.name == "Guido") | (Band.manager.name == "Graydon")
        ).run_sync()

        self.assertEqual(
            Band.count().where(Band.name == "New name").run_sync(), 2
        )



from unittest import TestCase

from piccolo.columns.column_types import (
    ForeignKey,
    LazyTableReference,
    Text,
    Varchar,
)
from piccolo.columns.m2m import M2M
from piccolo.schema import SchemaManager
from piccolo.table import Table
from tests.base import engines_skip

from .base import M2MBase


class Band(Table, schema="schema_1"):
    name = Varchar()
    genres = M2M(LazyTableReference("GenreToBand", module_path=__name__))


class Genre(Table, schema="schema_1"):
    name = Varchar()
    bands = M2M(LazyTableReference("GenreToBand", module_path=__name__))


class GenreToBand(Table, schema="schema_1"):
    band = ForeignKey(Band)
    genre = ForeignKey(Genre)
    reason = Text(help_text="For testing additional columns on join tables.")


@engines_skip("sqlite")
class TestM2MWithSchema(M2MBase, TestCase):
    """
    Make sure that when the tables exist in a non-public schema, that M2M still
    works.
    """

    band = Band
    genre = Genre
    genre_to_band = GenreToBand
    all_tables = [Band, Genre, GenreToBand]

    def tearDown(self):
        SchemaManager().drop_schema(
            schema_name="schema_1", cascade=True
        ).run_sync()



import pytest

from piccolo.query.methods.delete import DeletionError
from tests.base import DBTestCase, engine_version_lt, is_running_sqlite
from tests.example_apps.music.tables import Band


class TestDelete(DBTestCase):
    def test_delete(self):
        self.insert_rows()

        Band.delete().where(Band.name == "CSharps").run_sync()

        response = Band.count().where(Band.name == "CSharps").run_sync()

        self.assertEqual(response, 0)

    @pytest.mark.skipif(
        is_running_sqlite() and engine_version_lt(3.35),
        reason="SQLite version not supported",
    )
    def test_delete_returning(self):
        """
        Make sure delete works with the `returning` clause.
        """

        self.insert_rows()

        response = (
            Band.delete()
            .where(Band.name == "CSharps")
            .returning(Band.name)
            .run_sync()
        )

        self.assertEqual(len(response), 1)
        self.assertEqual(response, [{"name": "CSharps"}])

    def test_validation(self):
        """
        Make sure you can't delete all the data without forcing it.
        """
        with self.assertRaises(DeletionError):
            Band.delete().run_sync()

        Band.delete(force=True).run_sync()



from __future__ import annotations

import contextvars
import logging
import pprint
import typing as t
from abc import ABCMeta, abstractmethod

from piccolo.querystring import QueryString
from piccolo.utils.sync import run_sync
from piccolo.utils.warnings import Level, colored_string, colored_warning

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.query.base import Query


logger = logging.getLogger(__name__)


class Batch:
    pass


TransactionClass = t.TypeVar("TransactionClass")


class Engine(t.Generic[TransactionClass], metaclass=ABCMeta):

    __slots__ = ("query_id",)

    def __init__(self):
        run_sync(self.check_version())
        run_sync(self.prep_database())
        self.query_id = 0

    @property
    @abstractmethod
    def engine_type(self) -> str:
        pass

    @property
    @abstractmethod
    def min_version_number(self) -> float:
        pass

    @abstractmethod
    async def get_version(self) -> float:
        pass

    @abstractmethod
    def get_version_sync(self) -> float:
        pass

    @abstractmethod
    async def prep_database(self):
        pass

    @abstractmethod
    async def batch(
        self,
        query: Query,
        batch_size: int = 100,
        node: t.Optional[str] = None,
    ) -> Batch:
        pass

    @abstractmethod
    async def run_querystring(self, querystring: QueryString, in_pool: bool):
        pass

    @abstractmethod
    async def run_ddl(self, ddl: str, in_pool: bool = True):
        pass

    @abstractmethod
    def transaction(self):
        pass

    @abstractmethod
    def atomic(self):
        pass

    async def check_version(self):
        """
        Warn if the database version isn't supported.
        """
        try:
            version_number = await self.get_version()
        except Exception as exception:
            colored_warning(
                f"Unable to fetch server version: {exception}",
                level=Level.high,
            )
            return

        engine_type = self.engine_type.capitalize()
        logger.info(f"Running {engine_type} version {version_number}")
        if version_number and (version_number < self.min_version_number):
            message = (
                f"This version of {self.engine_type} isn't supported "
                f"(< {self.min_version_number}) - some features might not be "
                "available. For instructions on installing databases, see the "
                "Piccolo docs."
            )
            colored_warning(message, stacklevel=3)

    def _connection_pool_warning(self):
        message = (
            f"Connection pooling is not supported for {self.engine_type}."
        )
        logger.warning(message)
        colored_warning(message, stacklevel=3)

    async def start_connection_pool(self):
        """
        The database driver doesn't implement connection pooling.
        """
        self._connection_pool_warning()

    async def close_connection_pool(self):
        """
        The database driver doesn't implement connection pooling.
        """
        self._connection_pool_warning()

    ###########################################################################

    current_transaction: contextvars.ContextVar[t.Optional[TransactionClass]]

    def transaction_exists(self) -> bool:
        """
        Find out if a transaction is currently active.

        :returns:
            ``True`` if a transaction is already active for the current
            asyncio task. This is useful to know, because nested transactions
            aren't currently supported, so you can check if an existing
            transaction is already active, before creating a new one.

        """
        return self.current_transaction.get() is not None

    ###########################################################################
    # Logging queries and responses

    def get_query_id(self) -> int:
        self.query_id += 1
        return self.query_id

    def print_query(self, query_id: int, query: str):
        print(colored_string(f"\nQuery {query_id}:"))
        print(query)

    def print_response(self, query_id: int, response: t.List):
        print(
            colored_string(f"\nQuery {query_id} response:", level=Level.high)
        )
        pprint.pprint(response)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import os
import typing as t

from setuptools import find_packages, setup

from piccolo import __VERSION__ as VERSION

directory = os.path.abspath(os.path.dirname(__file__))

extras = ["orjson", "playground", "postgres", "sqlite", "uvloop"]


with open(os.path.join(directory, "README.md")) as f:
    LONG_DESCRIPTION = f.read()


def parse_requirement(req_path: str) -> t.List[str]:
    """
    Parse requirement file.
    Example:
        parse_requirement('requirements.txt')       # requirements/requirements.txt
        parse_requirement('extras/playground.txt')  # requirements/extras/playground.txt
    Returns:
        List[str]: list of requirements specified in the file.
    """  # noqa: E501
    with open(os.path.join(directory, "requirements", req_path)) as f:
        contents = f.read()
        return [i.strip() for i in contents.strip().split("\n")]


def extras_require() -> t.Dict[str, t.List[str]]:
    """
    Parse requirements in requirements/extras directory
    """
    extra_requirements = {
        extra: parse_requirement(os.path.join("extras", f"{extra}.txt"))
        for extra in extras
    }

    extra_requirements["all"] = list(
        itertools.chain.from_iterable(extra_requirements.values())
    )

    return extra_requirements


setup(
    name="piccolo",
    version=VERSION,
    description=(
        "A fast, user friendly ORM and query builder which supports asyncio."
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Daniel Townsend",
    author_email="dan@dantownsend.co.uk",
    python_requires=">=3.7.0",
    url="https://github.com/piccolo-orm/piccolo",
    packages=find_packages(exclude=("tests",)),
    package_data={
        "": [
            "templates/*",
            "templates/**/*",
            "templates/**/**/*",
            "templates/**/**/**/*",
        ],
        "piccolo": ["py.typed"],
    },
    project_urls={
        "Documentation": (
            "https://piccolo-orm.readthedocs.io/en/latest/index.html"
        ),
        "Source": "https://github.com/piccolo-orm/piccolo",
        "Tracker": "https://github.com/piccolo-orm/piccolo/issues",
    },
    install_requires=parse_requirement("requirements.txt"),
    extras_require=extras_require(),
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Framework :: AsyncIO",
        "Typing :: Typed",
        "Topic :: Database",
    ],
    entry_points={"console_scripts": ["piccolo = piccolo.main:main"]},
)



from tests.base import DBTestCase, engines_only, sqlite_only
from tests.example_apps.music.tables import Band, Manager


class TestGetAll(DBTestCase):
    def test_get_all(self):
        self.insert_row()

        response = Band.objects().run_sync()

        self.assertEqual(len(response), 1)

        instance = response[0]

        self.assertIsInstance(instance, Band)
        self.assertEqual(instance.name, "Pythonistas")

        # Now try changing the value and saving it.
        instance.name = "Rustaceans"
        save_query = instance.save()
        save_query.run_sync()

        self.assertEqual(
            Band.select(Band.name).output(as_list=True).run_sync()[0],
            "Rustaceans",
        )


class TestOffset(DBTestCase):
    @engines_only("postgres", "cockroach")
    def test_offset_postgres(self):
        """
        Postgres can do an offset without a limit clause.
        """
        self.insert_rows()

        response = Band.objects().order_by(Band.name).offset(1).run_sync()

        self.assertEqual(
            [i.name for i in response],
            ["Pythonistas", "Rustaceans"],
        )

    @sqlite_only
    def test_offset_sqlite(self):
        """
        SQLite requires a limit clause for offset to work.
        """
        self.insert_rows()
        query = Band.objects().order_by(Band.name).offset(1)

        with self.assertRaises(ValueError):
            query.run_sync()

        query = query.limit(5)

        response = query.run_sync()

        self.assertEqual(
            [i.name for i in response], ["Pythonistas", "Rustaceans"]
        )


class TestGet(DBTestCase):
    def test_get(self):
        self.insert_row()

        band = Band.objects().get(Band.name == "Pythonistas").run_sync()

        self.assertEqual(band.name, "Pythonistas")

    def test_get_prefetch(self):
        self.insert_rows()

        # With prefetch clause
        band = (
            Band.objects()
            .get((Band.name == "Pythonistas"))
            .prefetch(Band.manager)
            .run_sync()
        )
        self.assertIsInstance(band.manager, Manager)

        # Just passing it straight into objects
        band = (
            Band.objects(Band.manager)
            .get((Band.name == "Pythonistas"))
            .run_sync()
        )
        self.assertIsInstance(band.manager, Manager)


class TestGetOrCreate(DBTestCase):
    def test_simple_where_clause(self):
        """
        Make sure `get_or_create` works for simple where clauses.
        """
        # When the row doesn't exist in the db:
        Band.objects().get_or_create(
            Band.name == "Pink Floyd", defaults={"popularity": 100}
        ).run_sync()

        instance = (
            Band.objects().where(Band.name == "Pink Floyd").first().run_sync()
        )

        self.assertIsInstance(instance, Band)
        self.assertEqual(instance.name, "Pink Floyd")
        self.assertEqual(instance.popularity, 100)

        # When the row already exists in the db:
        Band.objects().get_or_create(
            Band.name == "Pink Floyd", defaults={Band.popularity: 100}
        ).run_sync()

        instance = (
            Band.objects().where(Band.name == "Pink Floyd").first().run_sync()
        )

        self.assertIsInstance(instance, Band)
        self.assertEqual(instance.name, "Pink Floyd")
        self.assertEqual(instance.popularity, 100)

    def test_complex_where_clause(self):
        """
        Make sure `get_or_create` works with complex where clauses.
        """
        self.insert_rows()

        # When the row already exists in the db:
        instance = (
            Band.objects()
            .get_or_create(
                (Band.name == "Pythonistas") & (Band.popularity == 1000)
            )
            .run_sync()
        )
        self.assertIsInstance(instance, Band)
        self.assertEqual(instance._was_created, False)

        # When the row doesn't exist in the db:
        instance = (
            Band.objects()
            .get_or_create(
                (Band.name == "Pythonistas2") & (Band.popularity == 2000)
            )
            .run_sync()
        )
        self.assertIsInstance(instance, Band)
        self.assertEqual(instance._was_created, True)

    def test_very_complex_where_clause(self):
        """
        Make sure `get_or_create` works with very complex where clauses.
        """
        self.insert_rows()

        # When the row already exists in the db:
        instance = (
            Band.objects()
            .get_or_create(
                (Band.name == "Pythonistas")
                & (Band.popularity > 0)
                & (Band.popularity < 5000)
            )
            .run_sync()
        )
        self.assertIsInstance(instance, Band)
        self.assertEqual(instance._was_created, False)

        # When the row doesn't exist in the db:
        instance = (
            Band.objects()
            .get_or_create(
                (Band.name == "Pythonistas2")
                & (Band.popularity > 10)
                & (Band.popularity < 5000)
            )
            .run_sync()
        )
        self.assertIsInstance(instance, Band)
        self.assertEqual(instance._was_created, True)

        # The values in the > and < should be ignored, and the default should
        # be used for the column.
        self.assertEqual(instance.popularity, 0)

    def test_joins(self):
        """
        Make sure that that `get_or_create` creates rows correctly when using
        joins.
        """
        instance = (
            Band.objects()
            .get_or_create(
                (Band.name == "My new band")
                & (Band.manager.name == "Excellent manager")
            )
            .run_sync()
        )
        self.assertIsInstance(instance, Band)
        self.assertEqual(instance._was_created, True)

        # We want to make sure the band name isn't 'Excellent manager' by
        # mistake.
        self.assertEqual(Band.name, "My new band")

    def test_prefetch_existing_object(self):
        """
        Make sure that that `get_or_create` works with the `prefetch` clause,
        when it's an existing row in the database.
        """
        self.insert_rows()

        # With prefetch clause
        band = (
            Band.objects()
            .get_or_create((Band.name == "Pythonistas"))
            .prefetch(Band.manager)
            .run_sync()
        )
        self.assertIsInstance(band.manager, Manager)
        self.assertEqual(band.manager.name, "Guido")

        # Just passing it straight into objects
        band = (
            Band.objects(Band.manager)
            .get_or_create((Band.name == "Pythonistas"))
            .run_sync()
        )
        self.assertIsInstance(band.manager, Manager)
        self.assertEqual(band.manager.name, "Guido")

    def test_prefetch_new_object(self):
        """
        Make sure that that `get_or_create` works with the `prefetch` clause,
        when the row is being created in the database.
        """
        manager = Manager({Manager.name: "Guido"})
        manager.save().run_sync()

        # With prefetch clause
        band = (
            Band.objects()
            .get_or_create(
                (Band.name == "New Band") & (Band.manager == manager)
            )
            .prefetch(Band.manager)
            .run_sync()
        )
        self.assertIsInstance(band.manager, Manager)
        self.assertEqual(band.name, "New Band")

        # Just passing it straight into objects
        band = (
            Band.objects(Band.manager)
            .get_or_create(
                (Band.name == "New Band 2") & (Band.manager == manager)
            )
            .run_sync()
        )
        self.assertIsInstance(band.manager, Manager)
        self.assertEqual(band.name, "New Band 2")
        self.assertEqual(band.manager.name, "Guido")



from tests.base import DBTestCase, engine_is
from tests.example_apps.music.tables import Band, Manager


class TestToDict(DBTestCase):
    def test_to_dict(self):
        """
        Make sure that `to_dict` works correctly.
        """
        self.insert_row()

        instance = Manager.objects().first().run_sync()
        dictionary = instance.to_dict()
        if engine_is("cockroach"):
            self.assertDictEqual(
                dictionary, {"id": dictionary["id"], "name": "Guido"}
            )
        else:
            self.assertDictEqual(dictionary, {"id": 1, "name": "Guido"})

    def test_nested(self):
        """
        Make sure that `to_dict` works correctly, when the object contains
        nested objects.
        """
        self.insert_row()

        instance = Band.objects(Band.manager).first().run_sync()
        dictionary = instance.to_dict()
        if engine_is("cockroach"):
            self.assertDictEqual(
                dictionary,
                {
                    "id": dictionary["id"],
                    "name": "Pythonistas",
                    "manager": {
                        "id": instance["manager"]["id"],
                        "name": "Guido",
                    },
                    "popularity": 1000,
                },
            )
        else:
            self.assertDictEqual(
                dictionary,
                {
                    "id": 1,
                    "name": "Pythonistas",
                    "manager": {"id": 1, "name": "Guido"},
                    "popularity": 1000,
                },
            )

    def test_filter_rows(self):
        """
        Make sure that `to_dict` works correctly with a subset of columns.
        """
        self.insert_row()

        instance = Manager.objects().first().run_sync()
        dictionary = instance.to_dict(Manager.name)
        self.assertDictEqual(dictionary, {"name": "Guido"})

    def test_nested_filter(self):
        """
        Make sure that `to_dict` works correctly with nested objects and
        filtering.
        """
        self.insert_row()

        instance = Band.objects(Band.manager).first().run_sync()
        dictionary = instance.to_dict(Band.name, Band.manager.id)
        if engine_is("cockroach"):
            self.assertDictEqual(
                dictionary,
                {
                    "name": "Pythonistas",
                    "manager": {"id": dictionary["manager"]["id"]},
                },
            )
        else:
            self.assertDictEqual(
                dictionary,
                {
                    "name": "Pythonistas",
                    "manager": {"id": 1},
                },
            )

    def test_aliases(self):
        """
        Make sure that `to_dict` works correctly with aliases.
        """
        self.insert_row()

        instance = Manager.objects().first().run_sync()
        dictionary = instance.to_dict(
            Manager.id, Manager.name.as_alias("title")
        )
        if engine_is("cockroach"):
            self.assertDictEqual(
                dictionary, {"id": dictionary["id"], "title": "Guido"}
            )
        else:
            self.assertDictEqual(dictionary, {"id": 1, "title": "Guido"})



from __future__ import annotations

import typing as t
from dataclasses import dataclass


@dataclass
class Choice:
    """
    When defining enums for ``Column`` choices, they can either be defined
    like:

    .. code-block:: python

        class Title(Enum):
            mr = 1
            mrs = 2

    If using Piccolo Admin, the values shown will be ``Mr`` and ``Mrs``. If you
    want more control, you can use ``Choice`` for the value instead.

    .. code-block:: python

        class Title(Enum):
            mr = Choice(value=1, display_name="Mr.")
            mrs = Choice(value=1, display_name="Mrs.")

    Now the values shown will be ``Mr.`` and ``Mrs.``.

    """

    value: t.Any
    display_name: str



from unittest import TestCase

from piccolo.query.mixins import OrderByDelegate


class TestOrderByDelegate(TestCase):
    def test_no_columns(self):
        """
        An exception should be raised if no columns are passed in.
        """
        delegate = OrderByDelegate()

        with self.assertRaises(ValueError) as manager:
            delegate.order_by()

        self.assertEqual(
            manager.exception.__str__(),
            "At least one column must be passed to order_by.",
        )
